/*
 * Copyright (c) 2018 by Intinor AB. All rights reserved.
 *
 * This file is part of "libRaptorQ".
 *
 * libRaptorQ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * libRaptorQ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * and a copy of the GNU Lesser General Public License
 * along with libRaptorQ.  If not, see <http://www.gnu.org/licenses/>.
 */


#pragma once

#include "RaptorQ/v1/common.hpp"
#include "RaptorQ/v1/Octet.hpp"
#include "RaptorQ/v1/multiplication.hpp"
#include "RaptorQ/v1/Matrix_Block.hpp"
#include "RaptorQ/v1/util/Matrix_Row.hpp"
#include <cmath>
#include <vector>
#include <stdlib.h>
#include <cstdlib>
#include <vector>
#include <iostream>
#include <Eigen/Dense>
#include <Eigen/Sparse>

namespace RaptorQ__v1 {
namespace Impl {

using DenseMtx = Eigen::Matrix<Octet, Eigen::Dynamic, Eigen::Dynamic,
                                                            Eigen::RowMajor>;
using SparseMtx = Eigen::SparseMatrix<Octet, Eigen::RowMajor>;
using PermVec = std::vector<uint16_t>;

// Align memory access for SSSE3/AVX2 instuctions
const uint32_t alignment = 32;

class RAPTORQ_LOCAL Matrix
{
public:
    Matrix() {};

    Matrix(uint32_t rows, uint32_t cols)
    {
        init(rows, cols);
    }

    Matrix(const Matrix& mtx)
    {
        init(mtx.row_count, mtx.column_count);
        memcpy(data, mtx.data, mtx.row_count * mtx.stride * sizeof(uint8_t));
    }

    Matrix(const DenseMtx& mtx)
    {
        init(mtx.rows(), mtx.cols());
        for (uint32_t i = 0; i < row_count; i++) {
            for (uint32_t j = 0; j < column_count; j++) {
                data[i * stride + j] = (uint8_t)mtx(i, j);
            }
        }
    }

    Matrix& operator= (const Matrix &mtx)
    {
        Matrix temp(mtx);
        swap(*this, temp);

        return *this;
    };

    Matrix(Matrix&& mtx) : Matrix()
    {
        swap(*this, mtx);
    }

    Matrix& operator= (Matrix&& mtx)
    {
        swap(*this, mtx);
        return *this;
    }

    ~Matrix()
    {
        free(data);
    }

    uint8_t& operator()(int32_t row, int32_t col)
    {
        assert((int)row_count >= row);
        assert((int)column_count >= col);
        return data[row * stride + col];
    }

    uint8_t operator()(int32_t row, int32_t col) const
    {
        assert((int)row_count >= row);
        assert((int)column_count >= col);
        return data[row * stride + col];
    }

    bool operator== (const Matrix& mtx) const
    {
        if (rows() != mtx.rows() || cols() != mtx.cols()) {
            return false;
        }
        for (int32_t i = 0; i < rows(); i++) {
            if (memcmp(data + i * stride,
                                    mtx.data + i * mtx.stride, cols()) != 0) {
                return false;
            }
        }
        return true;
    }

    int32_t rows() const { return (int)this->row_count; }
    int32_t cols() const { return (int)this->column_count; }

    void row_add(uint16_t lhs_row, const Matrix &rhs, uint16_t rhs_row)
    {
        assert(stride == rhs.stride);
        uint8_t *dest = data + lhs_row * stride;
        uint8_t *src = rhs.data + rhs_row * rhs.stride;
        Matrix_Row::row_add(dest, src, rhs.stride);
    }

    void row_div(uint16_t row, uint8_t scalar)
    {
        uint8_t *dest = data + row * stride;
        Matrix_Row::row_div(dest, scalar, stride);
    }

    void row_multiply_add(uint16_t lhs_row, Matrix &rhs, uint16_t rhs_row,
                                                                uint8_t scalar)
    {
        assert(stride == rhs.stride);
        if (scalar == 0) {
            return;
        }
        uint8_t *dest = data + lhs_row * stride;
        uint8_t *src = rhs.data + rhs_row * rhs.stride;
        Matrix_Row::row_multiply_add(dest, src, scalar, rhs.stride);
    }

    void row_multiply_sub(uint16_t lhs_row, Matrix &rhs, uint16_t rhs_row,
                                                                uint8_t scalar)
    {
        row_multiply_add(lhs_row, rhs, rhs_row, scalar);
    }

    void row_swap(uint16_t row1, uint16_t row2)
    {
        if (row1 == row2)
            return;
        uint8_t *temp = data + rows() * stride;
        uint8_t *r1 = data + row1 * stride;
        uint8_t *r2 = data + row2 * stride;
        memcpy(temp, r1,   stride);
        memcpy(r1,   r2,   stride);
        memcpy(r2,   temp, stride);
    }

    void col_swap(uint16_t col1, uint16_t col2)
    {
        if (col1 == col2)
            return;

        uint8_t temp;
        for (int32_t row = 0; row < rows(); row++) {
            if (data[row * stride + col1] != 0 ||
                                            data[row * stride + col2] != 0) {
                temp = data[row * stride + col1];
                data[row * stride + col1] = data[row * stride + col2];
                data[row * stride + col2] = temp;
            }
        }
    }

    void inline row_assign(uint16_t lhs_row, const Matrix &rhs,
                                                            uint16_t rhs_row)
    {
        assert(cols() == rhs.cols());
        uint8_t *dest = data + lhs_row * stride;
        uint8_t *src = rhs.data + rhs_row * rhs.stride;
        memcpy(dest, src, cols());
    }

    void conservative_resize(int32_t rows, int32_t cols)
    {
        assert(rows > 0);
        assert(cols > 0);
        assert(rows <= this->row_count);
        assert(cols <= this->column_count);
        this->row_count = rows;
        this->column_count = cols;
    }

    void set_identity ()
    {
        assert( cols() == rows() );
        memset(data, 0, rows() * stride);
        for (int32_t i = 0; i < rows(); i++) {
            data[i * stride + i] = 1;
        }
    }

    DenseMtx to_eigen()
    {
        uint8_t *dest, *src;
        DenseMtx ret = DenseMtx(rows(), cols());
        for (int32_t i = 0; i < rows(); i++) {
            src = data + i * stride;
            dest = (uint8_t *)ret.data() + i * cols();
            memcpy(dest, src, cols() * sizeof(uint8_t));
        }
        return ret;
    }

    Matrix_Block block(uint32_t start_row, uint32_t start_col,
                                    uint32_t block_rows, uint32_t block_cols)
    {

        Matrix_Block ret = Matrix_Block(data, start_row, start_col, block_rows,
                                                        block_cols, stride);
        return ret;
    }

    DenseMtx to_eigen(int32_t row, int32_t col, int32_t rows, int32_t cols)
    {
        uint8_t *dest, *src;
        DenseMtx ret = DenseMtx(rows, cols);
        for (int32_t i = row; i < row + rows; i++) {
            src = data + i * stride + col;
            dest = (uint8_t *)ret.data() + (i - row) * cols;
            memcpy(dest, src, cols * sizeof(uint8_t));
        }
        return ret;
    }

    void values_from_eigen(const DenseMtx& mtx, int32_t row, int32_t col,
                                                  int32_t rows, int32_t cols)
    {
        uint8_t *dest, *src;
        for (int32_t i = row; i < row + rows; i++) {
            dest = data + i * stride + col;
            src = (uint8_t *)mtx.data() + (i - row) * mtx.cols();
            memcpy(dest, src, cols * sizeof(uint8_t));
        }
    }

    void values_from_matrix(const Matrix& mtx, int32_t row,
                                    int32_t col, int32_t rows, int32_t cols)
    {
        uint8_t *dest, *src;
        for (int32_t i = row; i < row + rows; i++) {
            dest = data + i * stride + col;
            src = mtx.data + (i - row) * mtx.stride;
            memcpy(dest, src, cols * sizeof(uint8_t));
        }
    }

    void values_from_eigen(const DenseMtx& mtx)
    {
        assert( mtx.rows() == rows() );
        assert( mtx.cols() == cols() );
        for (int32_t i = 0; i < rows(); i++) {
            memcpy(data + i * stride, mtx.data() + i * cols(), cols() *
                                                            sizeof(uint8_t));
        }
    }

    void values_from_eigen_top_left(const SparseMtx& mtx)
    {
        for (int32_t i = 0; i < mtx.rows(); i++) {
            memset(data + i * stride, 0, mtx.cols());
        }
        for (int32_t k=0; k<mtx.outerSize(); ++k) {
            for (SparseMtx::InnerIterator it(mtx,k); it; ++it)
            {
                data[it.row() * stride + it.col()] = (uint8_t)it.value();
            }
        }
    }

    SparseMtx top_left_sparse_view(int32_t rows, int32_t cols) const
    {
        std::vector<Eigen::Triplet<Octet>> tripletList;
        tripletList.reserve((rows * cols) / 10);

        for (int32_t i = 0; i < rows; i++) {
            for (int32_t j = 0; j < cols; j++) {
                if (data[i * stride + j] != 0) {
                    tripletList.emplace_back(Eigen::Triplet<Octet>(i, j,
                                                Octet(data[i * stride + j])));
                }
            }
        }
        SparseMtx ret = SparseMtx(rows, cols);
        ret.setFromTriplets(tripletList.begin(), tripletList.end());
        return ret;
    }

    // Create a sparse matrix with custom row and column order
    SparseMtx permutated_top_left_sparse_view(int32_t rows, int32_t cols,
                                                PermVec &r, PermVec &c) const
    {
        std::vector<uint16_t> rp(r.size());
        std::vector<uint16_t> cp(c.size());
        for (uint32_t i = 0; i < rp.size();i++)
            rp[r[i]] = i;
        for (uint32_t i = 0; i < cp.size();i++)
            cp[c[i]] = i;

        std::vector<Eigen::Triplet<uint8_t>> tripletList;
        tripletList.reserve((rows * cols) / 10);
        for (int32_t i = 0; i < this->rows(); i++) {
            for (int32_t j = 0; j < this->cols(); j++) {
                if (data[i * stride + j] != 0) {
                    if (rp[i] < rows && cp[j] < cols) {
                        tripletList.emplace_back(Eigen::Triplet<uint8_t>(rp[i],
                                                cp[j], data[i * stride + j]));
                    }
                }
            }
        }
        SparseMtx ret = SparseMtx(rows, cols);
        ret.setFromTriplets(tripletList.begin(), tripletList.end());
        return ret;
    }

    // TODO: Make faster with SIMD and composite fields
    friend Matrix operator* (const Matrix &lhs, const Matrix &rhs)
    {
        assert(lhs.column_count == rhs.row_count);

        int32_t m = lhs.row_count;
        int32_t n = lhs.column_count;
        int32_t p = rhs.column_count;

        Matrix res = Matrix(m, p);

        uint16_t *a = lhs.get_copy(); // No stride
        uint16_t *b = rhs.get_transposed(); // No stride
        uint8_t *c = res.data; // Stride

        /* Pre calculate oct_log_no_if for a */
        for (int32_t i = 0; i < m * n;i++) {
            a[i] = oct_log_no_if[a[i]];
        }

        /* Pre calculate oct_log_no_if for b */
        for (int32_t i = 0; i < n * p;i++) {
            b[i] = oct_log_no_if[b[i]];
        }

        for (int32_t i = 0; i < m; i++) {
            for (int32_t j = 0; j < p; j++) {
                uint8_t dot = 0;
                for (int32_t k = 0; k < n; k++) {
                    dot = dot ^ oct_exp_no_if[a[i * n + k] + b[j * n + k]];
                }
                c[i * res.stride + j] = dot;
            }
        }

        free(a);
        free(b);
        return res;
    }

    uint32_t non_zeros_in_row(int32_t row, uint32_t start, uint32_t stop) const
    {
        uint8_t *row_data = data + row * stride;
        return Matrix_Row::row_non_zero(row_data, start, stop);
    }

    uint32_t first_non_zero_in_row(int32_t row, uint32_t start,
                                                        uint32_t stop) const
    {
        uint8_t *row_data = data + row * stride;
        return Matrix_Row::row_first_non_zero(row_data, start, stop);
        return 0;
    }

    uint32_t non_zeros() const
    {
        uint32_t non_zero_count = 0;

        for (uint32_t i = 0; i < row_count; i++) {
            for (uint32_t j = 0; j < column_count; j++) {
                if (data[i * stride + j] != 0)
                    non_zero_count++;
            }
        }

        return non_zero_count;
    }

private:
    uint8_t *data {NULL};
    uint32_t row_count {0};
    uint32_t column_count {0};
    uint32_t stride {0};

    void init (uint32_t row, uint32_t col)
    {
        this->stride = ((col + alignment-1) / alignment) * alignment;
        // The size is one extra row to use as temp during row_swap
        const size_t size = (row + 1) * this->stride * sizeof(uint8_t);

        uint8_t *d = NULL;
        if (!posix_memalign((void **) &d, alignment, size)) {
            memset(d, 0, size);

            this->data = d;
            this->row_count = row;
            this->column_count = col;
        }
    }

    void swap(Matrix &first, Matrix &second)
    {
        std::swap(first.data, second.data);
        std::swap(first.row_count, second.row_count);
        std::swap(first.column_count, second.column_count);
        std::swap(first.stride, second.stride);
    }

    // Get a transposed copy of the Matrix.
    uint16_t *get_transposed() const
    {
        uint16_t *copy = (uint16_t*)calloc(row_count * column_count,
                                                            sizeof(uint16_t));

        for (uint32_t i = 0; i < row_count; i ++ )
        {
            for (uint32_t j = 0; j < column_count; j ++ )
            {
                copy[j * row_count + i] = data[i * stride + j];
            }
        }
        return copy;
    }

    // Get a copy of the Matrix.
    uint16_t *get_copy() const
    {
        uint16_t *copy = (uint16_t*)calloc(row_count * column_count,
                                                            sizeof(uint16_t));

        for (uint32_t i = 0; i < row_count; i ++ )
        {
            for (uint32_t j = 0; j < column_count; j ++ )
            {
                copy[i * column_count + j] = data[i * stride + j];
            }
        }
        return copy;
    }
};
}
}
