/*
 * Copyright (c) 2016, Luca Fulchir<luca@fulchir.it>, All rights reserved.
 *
 * This file is part of "libRaptorQ".
 *
 * libRaptorQ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * libRaptorQ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * and a copy of the GNU Lesser General Public License
 * along with libRaptorQ.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "RaptorQ/v1/common.hpp"
#include "RaptorQ/v1/Operation.hpp"
#include <algorithm>
#include <atomic>
#include <cmath>
#include <deque>
#include <limits>
#include <memory>
#include <mutex>
#include <vector>
#include <deque>
#include <map>
#include <utility>


namespace RaptorQ__v1 {
namespace Impl {

std::vector<uint8_t> RAPTORQ_API ops_to_raw (const std::deque<Operation> &ops);
inline std::vector<uint8_t> RAPTORQ_API ops_to_raw (const std::deque<Operation>
                                                                        &ops)
{
    std::vector<uint8_t> ret;
    ret.reserve (ops.size() * 10);
    for (const auto &op : ops) {
        op.serialize(ret);
    }
    return ret;
}

std::deque<Operation> RAPTORQ_API raw_to_ops (const std::vector<uint8_t> &raw);
inline std::deque<Operation> RAPTORQ_API raw_to_ops (const std::vector<uint8_t>
                                                                        &raw)
{
    std::deque<Operation> ops;
    for(std::vector<uint8_t>::size_type i = 0; i < raw.size(); i++) {
        if (raw[i] == static_cast<uint8_t> (Operation::_t::SWAP)) {
            assert ((i + 4) < raw.size());
            uint16_t row_1 = Operation::deserialize_uint16(raw, i + 1);
            uint16_t row_2 = Operation::deserialize_uint16(raw, i + 3);
            i += 4;
            ops.emplace_back (Operation::_t::SWAP, row_1, row_2);
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::DIV)) {
            assert ((i + 3) < raw.size());
            uint16_t row_1 = Operation::deserialize_uint16(raw, i + 1);
            uint8_t scalar = raw[i + 3];
            i += 3;
            ops.emplace_back (Operation::_t::DIV, row_1, Octet(scalar));
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::ADD_MUL)) {
            assert ((i + 5) < raw.size());
            uint16_t row_1 = Operation::deserialize_uint16(raw, i + 1);
            uint16_t row_2 = Operation::deserialize_uint16(raw, i + 3);
            uint8_t scalar = raw[i + 5];
            i += 5;
            ops.emplace_back (Operation::_t::ADD_MUL, row_1, row_2,
                                                                Octet(scalar));
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::BLOCK)) {
            assert ((i + 8) < raw.size());
            uint16_t rows = Operation::deserialize_uint16(raw, i + 1);
            uint16_t cols = Operation::deserialize_uint16(raw, i + 3);
            uint32_t nnz  = Operation::deserialize_uint32(raw, i + 5);
            i += 8;

            assert ((i + nnz * 5) < raw.size());
            DenseMtx mtx = SparseMtx(rows, cols);
            for (uint32_t j = 0; j < nnz; j++) {
                uint16_t row = Operation::deserialize_uint16(raw, i + 1);
                uint16_t col = Operation::deserialize_uint16(raw, i + 3);
                mtx(row, col) = raw[i + 5];
                i += 5;
            }
            ops.emplace_back (Operation::_t::BLOCK, mtx);
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::REORDER)) {
            assert ((i + 4) < raw.size());
            uint32_t size  = Operation::deserialize_uint32(raw, i + 1);
            i += 4;

            assert ((i + size * 2) < raw.size());
            std::vector<uint16_t> vec;
            vec.reserve(size);
            for (uint32_t j = 0; j < size; j++) {
                uint16_t row = Operation::deserialize_uint16(raw, i + 1);
                vec.emplace_back(row);
                i += 2;
            }
            ops.emplace_back (Operation::_t::REORDER, vec);
        } else if (raw[i] == static_cast<uint8_t> (Operation::_t::NONE)) {
            ;
        } else {
            assert ("Invalid instruction!");
        }
    }
    return ops;
}

// easy compress from/to Matrix
std::vector<uint8_t> RAPTORQ_API mtx_to_raw (const Matrix &mtx);
inline std::vector<uint8_t> RAPTORQ_API mtx_to_raw (const Matrix &mtx)
{
    std::vector<uint8_t> ret;
    uint32_t rows = static_cast<uint32_t> (mtx.rows());
    uint32_t cols = static_cast<uint32_t> (mtx.cols());
    if (std::max(rows, cols) < 256) {
        ret.reserve (static_cast<size_t> (mtx.non_zeros() * 3));
        for (uint32_t row = 0; row < rows; ++row) {
            for (uint32_t col = 0; col < cols; ++col) {
                if (mtx (row, col) != 0) {
                    ret.emplace_back (row);
                    ret.emplace_back (col);
                    ret.emplace_back (mtx (row, col));
                }
            }
        }
    } else {
        ret.reserve (static_cast<size_t> (mtx.non_zeros() * 5));
        for (uint32_t row = 0; row < rows; ++row) {
            for (uint32_t col = 0; col < cols; ++col) {
                if (mtx (row, col) != 0) {
                    Operation::serialize_uint16(ret, row);
                    Operation::serialize_uint16(ret, col);
                    ret.emplace_back (mtx (row, col));
                }
            }
        }
    }
    return ret;
}

Matrix RAPTORQ_API raw_to_mtx (const std::vector<uint8_t> &raw,
                                    const uint32_t rows, const uint32_t cols);
inline Matrix RAPTORQ_API raw_to_mtx (const std::vector<uint8_t> &raw,
                                    const uint32_t rows, const uint32_t cols)
{
    Matrix ret (rows, cols);
    if (std::max(rows, cols) < 256) {
        for(std::vector<uint8_t>::size_type i = 0; i < raw.size(); i += 3) {
            uint32_t row = raw[i];
            uint32_t col = raw[i + 1];
            ret(row, col) = raw[i + 2];
        }
    } else {
        for(std::vector<uint8_t>::size_type i = 0; i < raw.size(); i += 5) {
            uint32_t row = Operation::deserialize_uint16(raw, i);
            uint32_t col = Operation::deserialize_uint16(raw, i + 2);
            ret(row, col) = raw[i + 4];
        }
    }
    return ret;
}

// Track cached data: matrices or operation vectors.
// we need to know the matrix size, repair overhead
// and if it's a matrix or an operations vector.
class RAPTORQ_API Cache_Key
{
public:
    Cache_Key (const uint16_t matrix_size, const uint32_t repair_overhead,
                    const bool ops_vec)
        :  _mt_size (matrix_size), _repair_overhead (repair_overhead),
                        _ops_vec(ops_vec)
    {}
    uint16_t _mt_size;
    uint32_t _repair_overhead;
    bool _ops_vec;

    bool operator< (const Cache_Key &rhs) const
    {
        if (_mt_size < rhs._mt_size)
            return true;
        if (_mt_size > rhs._mt_size)
            return false;
        // _mt_size == rhs._mt_size
        if (_repair_overhead < rhs._repair_overhead)
            return true;
        if (_repair_overhead > rhs._repair_overhead)
            return false;
        // _lost == rhs._lost

        if (rhs._ops_vec == false && _ops_vec == true) {
            return false;
        }
        if (rhs._ops_vec == true && _ops_vec == false) {
            return true;
        }
        // all the same
        return false;
    }
    bool operator== (const Cache_Key &rhs) const
    {
        return _mt_size == rhs._mt_size &&
                _repair_overhead == rhs._repair_overhead &&
                            _ops_vec == rhs._ops_vec;
    }
};


///////////////////////////////////////////////////////////////////////
// Decaying Least Frequency algorithm.
//   "Normal" LFU just tracks the most frequently used, so if an item
//   is used a lot at the beginning, and never again, it will still sit
//   at the top of the cache.
// The idea behind the "Decaying" LF is to add new elements at the back
// At every cache hit, the element is moved one step towards the back
// Elements are only removed from the front.
///////////////////////////////////////////////////////////////////////


// NOTE: no locking!! it is done in the memory section.

template<typename User_Data, typename Key>
class RAPTORQ_LOCAL DLF
{
public:
    DLF (const DLF&) = delete;
    DLF& operator= (const DLF&) = delete;
    DLF (DLF &&) = delete;
    DLF& operator= (DLF &&) = delete;

    static inline DLF *get()
    {
        // just drop the cache on exit time. Don't bother destroying
        // it correctly
        static DLF<User_Data, Key> *instance = new DLF<User_Data, Key>();
        return instance;
    }

    size_t get_size() const;
    size_t overhead_size() const;
    size_t resize (const size_t new_size);
    bool add (const Compress algo, User_Data &raw, const Key &key);
    std::pair<Compress, User_Data> get (const Key &key);
private:
    DLF ();
    // keep everything in a linked list, ordered by score.
    // later on we might more to a b-tree, keep it simple for now.
    // TODO: inefficient. All of this.
    // not even using "ghost" elements (tracking requests of non-cached elements
    // FIXME: test-only. Doesn't take into account overflow.

    class DLF_Data
    {
    public:
        DLF_Data (const DLF_Data &d);
        DLF_Data (const size_t _index, User_Data &_raw, const Compress alg)
            : index(_index), raw(std::move(_raw)), algorithm (alg)
        {}
        size_t index;
        User_Data raw;
        Compress algorithm;

        bool operator< (const DLF_Data &rhs) const;
        DLF_Data& operator= (const DLF_Data &rhs);
    };
    std::mutex biglock;
    size_t actual_size;
    size_t max_size;
    size_t removed;
    std::map<Key, DLF_Data> data;
    std::deque<Key> data_order;

    void move_towards_back(Key key);
};


template<typename User_Data, typename Key>
DLF<User_Data, Key>::DLF_Data::DLF_Data (const DLF_Data &d)
{
    algorithm = d.algorithm;
    index = d.index;
    raw = d.raw;
}

template<typename User_Data, typename Key>
typename DLF<User_Data, Key>::DLF_Data&
   DLF<User_Data, Key>::DLF_Data::operator= (const DLF_Data &d)
{
    algorithm = d.algorithm;
    index = d.index;
    raw = d.raw;

    return *this;
}

template<typename User_Data, typename Key>
DLF<User_Data, Key>::DLF()
{
    actual_size = 0;
    max_size = 0;
    removed = 0;
}

template<typename User_Data, typename Key>
size_t DLF<User_Data, Key>::get_size() const
    { return max_size; }

template<typename User_Data, typename Key>
size_t DLF<User_Data, Key>::overhead_size() const
    { return sizeof(DLF_Data) + 2 * sizeof(Key); }

template<typename User_Data, typename Key>
size_t DLF<User_Data, Key>::resize (const size_t new_size)
{
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED (guard);
    size_t overhead_size = sizeof(DLF_Data) + 2 * sizeof(Key);
    while (actual_size > new_size) {
        auto it = data_order.begin();
        assert (it != data_order.end() && "RQ: DLF: r_it should have data.");
        auto d_it = data.find(*it);
        actual_size -= (overhead_size + d_it->second.raw.size());
        assert ((d_it->second.index - removed) == 0 &&
                                                "RQ: DLF: index is not zero.");
        data.erase(d_it);
        data_order.pop_front();
        removed++;
    }
    max_size = new_size;
    return max_size;
}

template<typename User_Data, typename Key>
void DLF<User_Data, Key>::move_towards_back(Key key)
{
    if (data_order.size() <= 1)
        return; // Only one entry

    auto after = data.find(key);
    auto after_index = after->second.index - removed;
    assert (after_index >= 0 && after_index < (data_order.size()) &&
                                            "DLF: Index is out of bounds.");
    if (after_index == data_order.size() - 1)
        return; // Entry is already at the back

    auto before = data.find(data_order[after_index + 1]);
    auto before_index = before->second.index - removed;
    assert (before_index >= 0 && before_index < (data_order.size()) &&
                                            "DLF: Index is out of bounds.");
    assert (before_index != after_index && "DLF: Index is not unique.");
    // Swap keys
    std::swap(data_order[after_index], data_order[before_index]);
    // Swap index reference
    std::swap(after->second.index, before->second.index);
}

template<typename User_Data, typename Key>
std::pair<Compress, User_Data> DLF<User_Data, Key>::get (const Key &key)
{
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED(guard);
    auto value = data.find(key);
    if (value != data.end()) {
        std::pair<Compress, User_Data> ret_data = {value->second.algorithm,
                                                            value->second.raw};
        move_towards_back(key);
        return ret_data;
    }
    return {Compress::NONE, User_Data ()};
}

template<typename User_Data, typename Key>
bool DLF<User_Data, Key>::add (const Compress algorithm, User_Data &raw,
                                                                const Key &key)
{
    assert (data.size() == data_order.size() && "DLF: Size differ.");
    size_t raw_size = raw.size();
    std::lock_guard<std::mutex> guard (biglock);
    RQ_UNUSED(guard);
    auto value = data.find(key);
    if (value != data.end()) {
        move_towards_back(key);
        return true;
    }
    size_t overhead_size = sizeof(DLF_Data) + 2 * sizeof(Key);
    // key not present.
    if (max_size - actual_size >= overhead_size + raw_size) {
        // free space is the best
        DLF_Data dlf(removed + data_order.size(), raw, algorithm);
        data.insert(std::make_pair(key, std::move(dlf)));
        data_order.push_back(key);
        actual_size += overhead_size + raw_size;
        return true;
    } else {
        // need to delete some element?
        size_t usable_bytes = max_size - actual_size;
        size_t delete_from_begin = 0;
        for (auto it = data_order.begin(); it != data_order.end() &&
                                usable_bytes < (overhead_size + raw_size);
                                                                    ++it) {
            auto d_it = data.find(*it);
            usable_bytes +=  overhead_size + d_it->second.raw.size();
            ++delete_from_begin;
        }
        if (usable_bytes < (overhead_size + raw_size)) {
            // can't delete enough cached items, the new item requires
            // too much space, and fresher elements are present.
            return false;
        }
        while (delete_from_begin > 0) {
            auto it = data_order.begin();
            auto d_it = data.find(*it);
            actual_size -= (overhead_size + d_it->second.raw.size());
            assert ((d_it->second.index - removed) == 0 &&
                                                "RQ: DLF: index is not zero.");
            data.erase(d_it);
            data_order.pop_front();
            removed++;
            --delete_from_begin;
        }
        DLF_Data dlf(removed + data_order.size(), raw, algorithm);
        data.insert(std::make_pair(key, std::move(dlf)));
        data_order.push_back(key);
        actual_size = max_size - usable_bytes;
        actual_size += overhead_size + raw_size;
        return true;
    }
}


} // namespace Impl
} // namespace RaptorQ__v1
