/*
 * Copyright (c) 2018 by Intinor AB. All rights reserved.
 *
 * This file is part of "libRaptorQ".
 *
 * libRaptorQ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * libRaptorQ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * and a copy of the GNU Lesser General Public License
 * along with libRaptorQ.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <immintrin.h>
#include "RaptorQ/v1/gf256.hpp"
#include "RaptorQ/v1/multiplication.hpp"

namespace RaptorQ__v1 {
namespace Impl {
namespace Matrix_Row_SIMD {
    void add_ssse3(uint8_t *dest, uint8_t *src, int bytes)
    {
        uint8_t *src_ptr, *dest_ptr;
        __m128i s, d;

        assert(bytes % 16 == 0);
        assert(bytes >= 16);

        src_ptr = src;
        dest_ptr = dest;

        for (int i = 0; i < bytes; i += 16) {
            s = _mm_load_si128((__m128i *) (&src_ptr[i]));
            d = _mm_load_si128((__m128i *) (&dest_ptr[i]));
            d = _mm_xor_si128(d, s);
            _mm_store_si128((__m128i *) (&dest_ptr[i]), d);
        }
    }

    void div_ssse3(uint8_t *data, uint8_t num, int bytes)
    {
        uint8_t *high, *low, *data_ptr;
        __m128i mask, t, r, v, m_table_high, m_table_low;

        assert(bytes % 16 == 0);
        assert(bytes >= 16);

        // Inverse num (div is multiply with 1/num)
        num = oct_exp[255 - oct_log[num - 1]];

        high = gf256_high_ptr + (num << 4) * 2;
        low = gf256_low_ptr + (num << 4) * 2;

        mask = _mm_set1_epi8(0x0f);
        m_table_high = _mm_loadu_si128((__m128i *) (high));
        m_table_low = _mm_loadu_si128((__m128i *) (low));

        data_ptr = data;
        for (int i = 0; i < bytes; i += 16) {
            v = _mm_load_si128((__m128i *) (&data_ptr[i]));
            t = _mm_and_si128(mask, v);
            r = _mm_shuffle_epi8(m_table_low, t);
            v = _mm_srli_epi64(v, 4);
            t = _mm_and_si128(mask, v);
            r = _mm_xor_si128(r, _mm_shuffle_epi8(m_table_high, t));
            _mm_store_si128((__m128i *) (&data_ptr[i]), r);
        }
    }

    void multiply_and_add_ssse3(uint8_t *dest, uint8_t *src, uint8_t num,
                                                                    int bytes)
    {
        uint8_t *high, *low, *src_ptr, *dest_ptr;
        __m128i mask, t, r, v, m_table_high, m_table_low;

        assert(bytes % 16 == 0);
        assert(bytes >= 16);

        high = gf256_high_ptr + (num << 4) * 2;
        low = gf256_low_ptr + (num << 4) * 2;

        mask = _mm_set1_epi8(0x0f);
        m_table_high = _mm_loadu_si128((__m128i *) (high));
        m_table_low = _mm_loadu_si128((__m128i *) (low));

        src_ptr = src;
        dest_ptr = dest;
        for (int i = 0; i < bytes; i += 16) {
            v = _mm_load_si128((__m128i *) (&src_ptr[i]));
            t = _mm_and_si128(mask, v);
            r = _mm_shuffle_epi8(m_table_low, t);
            v = _mm_srli_epi64(v, 4);
            t = _mm_and_si128(mask, v);
            r = _mm_xor_si128(r, _mm_shuffle_epi8(m_table_high, t));
            v = _mm_load_si128((__m128i *) (&dest_ptr[i]));
            r = _mm_xor_si128(r, v);
            _mm_store_si128((__m128i *) (&dest_ptr[i]), r);
        }
    }

    uint32_t non_zero_ssse3(uint8_t *data, uint32_t start, uint32_t stop)
    {
        uint8_t *data_ptr;
        __m128i zero16, cmp, sum16;
        uint32_t non_zero_count = 0;

        // SIMD values
        uint32_t limit = ((stop - start) / 16);

        zero16 = _mm_setzero_si128();
        data_ptr = data + start;

        uint32_t count = 0;
        uint32_t loop_limit = limit;
        non_zero_count += limit*16;
        do {
            sum16 = _mm_setzero_si128();
            count = std::min(loop_limit, (uint32_t)255);
            for (uint32_t i = 0; i < count; i++) {
                cmp = _mm_loadu_si128((__m128i *) (data_ptr));
                cmp = _mm_cmpeq_epi8(cmp, zero16);
                sum16 = _mm_sub_epi8(sum16, cmp);
                data_ptr += 16;
            }
            // Horizontal sum of 16 bytes
            __m128i sum1 = _mm_sad_epu8(sum16,zero16);
            __m128i sum2 = _mm_shuffle_epi32(sum1,2);
            __m128i sum3 = _mm_add_epi16(sum1,sum2);
            non_zero_count -= _mm_cvtsi128_si32(sum3);
            loop_limit -= count;
        } while (loop_limit > 0);

        // Post SIMD values
        for (uint32_t j = start + limit * 16; j < stop; j++)
            non_zero_count += (data[j] != 0);

        return non_zero_count;
    }

    uint32_t first_non_zero_ssse3(uint8_t *data, uint32_t start, uint32_t stop)
    {
        uint8_t *data_ptr;
        __m128i zero16, cmp;

        // SIMD values
        uint32_t limit = ((stop - start) / 16) * 16;
        zero16 = _mm_setzero_si128();
        data_ptr = data + start;
        for (uint32_t i = 0; i < limit; i++) {
            cmp = _mm_loadu_si128((__m128i *)&data_ptr[i]);
            cmp = _mm_cmpeq_epi8(cmp, zero16);
            int mask = _mm_movemask_epi8(cmp);
            if(mask != 0xffff) {
                for (uint32_t j = i; j < i + 16; j++) {
                    if (data_ptr[j] != 0) {
                        return start + j;
                    }
                }
            }
        }

        // Post SIMD values
        for (uint32_t j = start + limit; j < stop; j++) {
            if (data[j] != 0) {
                return j;
            }
        }
        assert(0);
        return 0;
    }
} // namespace Matrix_Row_SIMD
} // namespace Impl
} // namespace RaptorQ__v1
