#include <RaptorQ/v1/Octet.hpp>
#include <RaptorQ/v1/Matrix.hpp>
#include <memory>
#include <utility>
#include <iostream>
#include <chrono>
#include <cstdlib>
#include "catch2/catch.hpp"

using namespace RaptorQ__v1::Impl;

TEST_CASE( "Permutate with matrices", "[Matrix]" )
{
    Matrix A = Matrix(4, 4);
    Matrix R = Matrix(4, 4);
    Matrix C = Matrix(4, 4);
    R.set_identity();
    C.set_identity();
    for (int i = 0; i < 4;i++) {
        for (int j = 0; j < 4;j++) {
            A(i,j) = i * 4 + j;
        }
    }
    Matrix A_copy = A;

    A.row_swap(1,3);
    R.row_swap(1,3);

    A.row_swap(2,3);
    R.row_swap(2,3);

    Matrix D = R * A_copy;
    REQUIRE( A.to_eigen() == D.to_eigen() );

    A.col_swap(1,3);
    C.col_swap(1,3);

    A.col_swap(2,3);
    C.col_swap(2,3);

    A.row_swap(0,3);
    R.row_swap(0,3);

    A.col_swap(2,3);
    C.col_swap(2,3);

    A.row_swap(0,3);
    R.row_swap(0,3);

    D = R * A_copy * C;
    REQUIRE( A.to_eigen() == D.to_eigen() );
}

TEST_CASE( "Permutate with vectors", "[Matrix]" )
{
    Matrix A = Matrix(4, 6);
    std::vector<uint16_t> r(4);
    std::vector<uint16_t> c(6);
    for (int i = 0; i < 4;i++)
        r[i] = i;
    for (int i = 0; i < 6;i++)
        c[i] = i;
    for (int i = 0; i < 4;i++) {
        for (int j = 0; j < 6;j++) {
            A(i,j) = i * 4 + j;
        }
    }

    Matrix A_copy = A;

    A.row_swap(1,3);
    std::swap(r[1], r[3]);

    SparseMtx B = A_copy.permutated_top_left_sparse_view(4, 6, r, c);
    REQUIRE( A.to_eigen() == DenseMtx(B) );

    A.row_swap(2,3);
    std::swap(r[2], r[3]);

    A.col_swap(1, 5);
    std::swap(c[1], c[5]);

    SparseMtx C = A_copy.permutated_top_left_sparse_view(4, 6, r, c);
    REQUIRE( A.to_eigen() == DenseMtx(C) );

    A.col_swap(4, 5);
    std::swap(c[4], c[5]);

    A.col_swap(2, 3);
    std::swap(c[2], c[3]);

    A.row_swap(0, 3);
    std::swap(r[0], r[3]);

    A.col_swap(2, 3);
    std::swap(c[2], c[3]);

    A.row_swap(0, 3);
    std::swap(r[0], r[3]);

    SparseMtx D = A_copy.permutated_top_left_sparse_view(4, 6, r, c);
    REQUIRE( A.to_eigen() == DenseMtx(D) );

    D = A_copy.permutated_top_left_sparse_view(3, 2, r, c);
    REQUIRE( A.to_eigen().block(0, 0, 3, 2) == DenseMtx(D) );
}

TEST_CASE( "Create, set and get dense", "[Matrix]" )
{
    Matrix A = Matrix(4, 4);

    // Get when empty
    REQUIRE( A(2, 3) == 0 );

    A(2, 3) = 45; // Set at position
    REQUIRE( A(2, 3) == 45 ); // Get at position
    REQUIRE( A.non_zeros() == 1 );

    A(2, 3) = 15;
    REQUIRE( A(2, 3) == 15 );
    REQUIRE( A.non_zeros() == 1 );

    A(3, 3) = 73;
    REQUIRE( A(3, 3) == 73 );
    REQUIRE( A.non_zeros() == 2 );

    A(2, 0) = 43;
    REQUIRE( A(2, 0) == 43 );
    REQUIRE( A.non_zeros() == 3 );

    A(0, 2) = 241;
    REQUIRE( A(0, 2) == 241 );
    REQUIRE( A.non_zeros() == 4 );

    REQUIRE( A(2, 3) == 15 );
    REQUIRE( A(3, 3) == 73 );
    REQUIRE( A(2, 0) == 43 );
    REQUIRE( A(0, 2) == 241 );

    A(1, 1) = 0;
    A(1, 2) = 0;
    REQUIRE( A.non_zeros() == 4 );
}

TEST_CASE( "Copy dense", "[Matrix]" )
{
    Matrix A = Matrix(4, 4);
    REQUIRE( A(2, 3) == 0 );

    A(2, 3) = 45;
    REQUIRE( A(2, 3) == 45 );
    REQUIRE( A.non_zeros() == 1 );

    Matrix B = A;
    REQUIRE( B(2, 3) == 45 );
    REQUIRE( B.non_zeros() == 1 );

    B(2, 3) = 74;
    REQUIRE( B(2, 3) == 74 );
    REQUIRE( A(2, 3) == 45 );
    REQUIRE( A.non_zeros() == 1 );

    Matrix C;
    C = Matrix(B);
    REQUIRE( C(2, 3) == 74 );
    REQUIRE( B(2, 3) == 74 );
    REQUIRE( A(2, 3) == 45 );
    REQUIRE( C.non_zeros() == 1 );

    C(0, 0) = 5;
    REQUIRE( A.non_zeros() == 1 );
    REQUIRE( B.non_zeros() == 1 );
    REQUIRE( C.non_zeros() == 2 );
}

TEST_CASE( "Multiply dense", "[Matrix]" )
{
    Matrix A = Matrix(2, 3);
    Matrix B = Matrix(3, 2);

    A(0, 0) = 1;
    A(0, 1) = 2;
    A(0, 2) = 3;
    A(1, 0) = 4;
    A(1, 1) = 5;
    A(1, 2) = 6;

    B(0, 0) = 7;
    B(0, 1) = 8;
    B(1, 0) = 9;
    B(1, 1) = 10;
    B(2, 0) = 11;
    B(2, 1) = 12;

    Matrix C = A * B;


    REQUIRE( C.rows() == 2 );
    REQUIRE( C.cols() == 2 );
    REQUIRE( C(0, 0) == 8 );
    REQUIRE( C(0, 1) == 8 );
    REQUIRE( C(1, 0) == 11 );
    REQUIRE( C(1, 1) == 42 );
    REQUIRE( C.non_zeros() == 4 );


    A = Matrix(3, 3);
    B = Matrix(3, 3);

    A(0, 1) = 2;
    A(0, 2) = 3;
    A(1, 0) = 4;
    A(2, 2) = 5;

    B(0, 2) = 7;
    B(1, 1) = 8;
    B(2, 1) = 9;
    B(2, 2) = 6;

    C = A * B;

    REQUIRE( C.rows() == 3 );
    REQUIRE( C.cols() == 3 );
    REQUIRE( C(0, 1) == 11 );
    REQUIRE( C(0, 2) == 10 );
    REQUIRE( C(1, 2) == 28 );
    REQUIRE( C(2, 1) == 45 );
    REQUIRE( C(2, 2) == 30 );
    REQUIRE( C.non_zeros() == 5 );
}

TEST_CASE( "Eigen conversion and verification", "[Matrix]" )
{
    DenseMtx A = DenseMtx(2, 3);
    DenseMtx B = DenseMtx(3, 2);

    A(0, 0) = 1;
    A(0, 1) = 2;
    A(0, 2) = 3;
    A(1, 0) = 4;
    A(1, 1) = 5;
    A(1, 2) = 6;

    B(0, 0) = 7;
    B(0, 1) = 8;
    B(1, 0) = 9;
    B(1, 1) = 10;
    B(2, 0) = 11;
    B(2, 1) = 12;

    DenseMtx C = A * B;

    Matrix D = Matrix(A);
    Matrix E = Matrix(B);
    Matrix F = D * E;

    REQUIRE( C(0, 0) == F(0, 0) );
    REQUIRE( C(0, 1) == F(0, 1) );
    REQUIRE( C(1, 0) == F(1, 0) );
    REQUIRE( C(1, 1) == F(1, 1) );

    DenseMtx G = F.to_eigen();
    REQUIRE( C == G );
}

TEST_CASE( "Eigen multiplication", "[Matrix]" )
{
    DenseMtx A = DenseMtx(10, 10);
    DenseMtx B = DenseMtx(10, 10);

    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = (i * j) % 256;
            B(i, j) = (3 * i * j) % 256;
        }
    }

    DenseMtx C = A * B;

    Matrix D = Matrix(A);
    Matrix E = Matrix(B);
    Matrix F = D * E;
    REQUIRE( C == F.to_eigen() );
}

TEST_CASE( "Eigen top left", "[Matrix]" )
{
    DenseMtx A = DenseMtx(10, 10);
    DenseMtx B = DenseMtx(10, 10);

    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 256;
            B(i, j) = (3 * i + 2 * j) % 256;
        }
    }

    DenseMtx C = A * B;
    auto Cb = C.block(0, 0, 5, 5);
    Cb = A.block(0, 0, 5, 5) * B.block(0, 0, 5, 5);

    Matrix D = Matrix(A);
    Matrix E = Matrix(B);
    Matrix F = D * E;

    SparseMtx Sd = D.top_left_sparse_view(5, 5);
    SparseMtx Se = E.top_left_sparse_view(5, 5);

    SparseMtx Sf = Sd * Se;

    REQUIRE( C != F.to_eigen() );
    F.values_from_eigen_top_left(Sf);

    REQUIRE( Cb == DenseMtx(Sf) );
    REQUIRE( C == F.to_eigen() );
}

TEST_CASE( "Eigen block", "[Matrix]" )
{
    // Basic test comparing Eigen to Matrix
    const int org_w = 15, org_h = 15;

    const int new_w = 5,  new_h = 5;
    const int x0    = 3,  y0    = 4;

    DenseMtx A = DenseMtx(org_w, org_w);
    for (int i = 0; i < org_w; i++) {
        for (int j = 0; j < org_h; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 256;
        }
    }

    Matrix B = Matrix(A);

    auto Ab = A.block(x0, y0, new_w, new_h);
    auto Bb = B.block(x0, y0, new_w, new_h);

    REQUIRE( Ab == Bb.to_eigen() );

}

#if 0
TEST_CASE( "Eigen block 2", "[Matrix]" )
{
    DenseMtx A = DenseMtx(10, 10);
    DenseMtx B = DenseMtx(10, 10);

    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 256;
            B(i, j) = (3 * i + 2 * j) % 256;
        }
    }

    DenseMtx C = A * B;
    auto Cb = C.block(0, 0, 5, 5);
    Cb = A.block(0, 0, 5, 5) * B.block(0, 0, 5, 5);

    Matrix D = Matrix(A);
    Matrix E = Matrix(B);
    Matrix F = D * E;

    REQUIRE( C != F.to_eigen() );

    auto Fb = F.block(0,0,5,5);
    for (int i = 0; i < 5; i++) {
        for (int j = 0; j < 5; j++) {
            Fb(i, j) = Cb(i, j);
        }
    }

    REQUIRE( Cb == DenseMtx(Fb) );
    REQUIRE( C == F.to_eigen() );
}
#endif

TEST_CASE( "Set zero block matrix", "[Matrix]" )
{
    const int org_w = 15, org_h = 15;

    const int new_w = 5,  new_h = 5;
    const int x0    = 3,  y0    = 4;

    DenseMtx A = DenseMtx(org_w, org_w);
    for (int i = 0; i < org_w; i++) {
        for (int j = 0; j < org_h; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 255 + 1;
        }
    }

    Matrix B = Matrix(A);

    REQUIRE( A == B.to_eigen() );
    auto Ab = A.block(x0, y0, new_w, new_h);
    auto Bb = B.block(x0, y0, new_w, new_h);

    for (int i = x0; i < new_w; i++) {
        for (int j = x0; j < new_h; j++) {
            REQUIRE( A(i, j) != 0 );
            REQUIRE( B(i, j) != 0 );
        }
    }

    Ab.setZero();
    Bb.set_zero();

    for (int i = 0; i < new_w; i++) {
        for (int j = 0; j < new_h; j++) {
            REQUIRE( Ab(i, j) == 0 );
            REQUIRE( Bb(i, j) == 0 );
        }
    }

    for (int i = x0; i < x0 + new_w; i++) {
        for (int j = y0; j < y0 + new_h; j++) {
            REQUIRE( A(i, j) == 0 );
            REQUIRE( B(i, j) == 0 );
        }
    }

    REQUIRE( A == B.to_eigen() );
}

TEST_CASE( "Copy operations", "[Matrix]" )
{
    DenseMtx A = DenseMtx(10, 10);
    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 255 + 1;
        }
    }

    Matrix B = Matrix(A);

    REQUIRE( A == B.to_eigen() );

    Matrix C = Matrix(10, 10);
    Matrix D = Matrix(10, 10);
    DenseMtx F = DenseMtx(10, 10);
    F.setZero();

    C.values_from_eigen(A, 1, 2, 5, 4);
    D.values_from_matrix(B, 1, 2, 5, 4);
    F.block(1, 2, 5, 4) = A.block(0, 0, 5, 4);

    REQUIRE( A != C.to_eigen() );
    REQUIRE( F == C.to_eigen() );
    REQUIRE( C == D );

    C.values_from_eigen(A, 3, 1, 6, 8);
    D.values_from_matrix(B, 3, 1, 6, 8);
    F.block(3, 1, 6, 8) = A.block(0, 0, 6, 8);

    REQUIRE( A != C.to_eigen() );
    REQUIRE( F == C.to_eigen() );
    REQUIRE( C == D );

    C.values_from_eigen(A, 0, 0, 6, 8);
    D.values_from_matrix(B, 0, 0, 6, 8);
    F.block(0, 0, 6, 8) = A.block(0, 0, 6, 8);
    C.values_from_eigen(A, 4, 4, 6, 6);
    D.values_from_matrix(B, 4, 4, 6, 6);
    F.block(4, 4, 6, 6) = A.block(0, 0, 6, 6);

    REQUIRE( A != C.to_eigen() );
    REQUIRE( F == C.to_eigen() );
    REQUIRE( C == D );

    C.values_from_eigen(A, 0, 4, 10, 6);
    D.values_from_matrix(B, 0, 4, 10, 6);
    F.block(0, 4, 10, 6) = A.block(0, 0, 10, 6);
    C.values_from_eigen(A, 4, 0, 6, 10);
    D.values_from_matrix(B, 4, 0, 6, 10);
    F.block(4, 0, 6, 10) = A.block(0, 0, 6, 10);

    REQUIRE( A != C.to_eigen() );
    REQUIRE( F == C.to_eigen() );
    REQUIRE( C == D );

    C.values_from_eigen(A, 0, 0, 10, 10);
    D.values_from_matrix(B, 0, 0, 10, 10);
    F.block(0, 0, 10, 10) = A.block(0, 0, 10, 10);

    REQUIRE( A == C.to_eigen() );
    REQUIRE( F == C.to_eigen() );
    REQUIRE( B == C );
    REQUIRE( C == D );
}

TEST_CASE( "Value to/from Eigen", "[Matrix]" )
{
    DenseMtx A = DenseMtx(10, 10);
    DenseMtx B = DenseMtx(10, 10);
    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 255 + 1;
            B(i, j) = ((i + 7) * (j + 13)) % 255 + 1;
        }
    }

    Matrix C = Matrix(A);

    A.block (2, 3, 4, 5) = B.block (4, 2, 4, 5);
    REQUIRE( A != C.to_eigen() );

    C.values_from_eigen(B.block (4, 2, 4, 5), 2, 3, 4, 5);
    REQUIRE( A == C.to_eigen() );

    A.block (0, 0, 5, 7) = B.block (0, 0, 5, 7);
    REQUIRE( A != C.to_eigen() );

    C.values_from_eigen(B, 0, 0, 5, 7);
    REQUIRE( A == C.to_eigen() );

    REQUIRE( C.to_eigen(0, 0, 5, 7) == A.block (0, 0, 5, 7) );
    REQUIRE( C.to_eigen(1, 0, 5, 7) == A.block (1, 0, 5, 7) );
    REQUIRE( C.to_eigen(0, 2, 5, 7) == A.block (0, 2, 5, 7) );
    REQUIRE( C.to_eigen(4, 6, 3, 2) == A.block (4, 6, 3, 2) );
}

TEST_CASE( "Copy construct", "[Matrix]" )
{
    Matrix A = Matrix(10, 10);
    Matrix B = Matrix(10, 10);
    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 255 + 1;
        }
    }

    REQUIRE( A.to_eigen() != B.to_eigen() );
    B = Matrix(A);
    REQUIRE( A.to_eigen() == B.to_eigen() );
    B(0, 0) += 1;
    REQUIRE( A.to_eigen() != B.to_eigen() );
}

TEST_CASE( "Copy", "[Matrix]" )
{
    Matrix A = Matrix(10, 10);
    Matrix B = Matrix(10, 10);
    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 255 + 1;
        }
    }

    REQUIRE( A.to_eigen() != B.to_eigen() );
    B = A;
    REQUIRE( A.to_eigen() == B.to_eigen() );
    B(0, 0) += 1;
    REQUIRE( A.to_eigen() != B.to_eigen() );
}

TEST_CASE( "Move", "[Matrix]" )
{
    Matrix A = Matrix(10, 10);
    Matrix B = Matrix(10, 10);
    for (int i = 0; i < 10; i++) {
        for (int j = 0; j < 10; j++) {
            A(i, j) = ((i + 12) * (j + 7)) % 255 + 1;
        }
    }

    Matrix C = Matrix(A);
    REQUIRE( A.to_eigen() == C.to_eigen() );
    REQUIRE( A.to_eigen() != B.to_eigen() );
    B = std::move(A);
    REQUIRE( B.to_eigen() == C.to_eigen() );
    B(0, 0) += 1;
    REQUIRE( B.to_eigen() != C.to_eigen() );
}
